#include "TelnetLooper.h"
#include "TelnetMessages.h"

#include <TCPSockets.h>
#include <support/Beep.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

inline void TelnetLooper::SendOption(option_t option, int state)
{
	// TODO: Debugging stuff, clean me out when you can.
	fprintf(stderr, "Send: <IAC>");
	switch (state) {
	case TELNET_WILL:
		fprintf(stderr, "<WILL>");
		break;
	case TELNET_WONT:
		fprintf(stderr, "<WONT>");
		break;
	case TELNET_DO:
		fprintf(stderr, "<DO>");
		break;
	case TELNET_DONT:
		fprintf(stderr, "<DONT>");
		break;
	default:
		fprintf(stderr, "<%d>", state);
	}
	switch (option) {
	case XMIT_BIN:
		fprintf(stderr, "<XMIT_BIN>");
		break;
	case ECHO:
		fprintf(stderr, "<ECHO>");
		break;
	case NO_GA:
		fprintf(stderr, "<NO_GA>");
		break;
	case STATUS:
		fprintf(stderr, "<STATUS>");
		break;
	case TM:
		fprintf(stderr, "<TM>");
		break;
	case TERM:
		fprintf(stderr, "<TERM>");
		break;
	case EOR:
		fprintf(stderr, "<EOR>");
		break;
	case NAWS:
		fprintf(stderr, "<NAWS>");
		break;
	case TTYSPEED:
		fprintf(stderr, "<TTYSPEED>");
		break;
	case RFLOW:
		fprintf(stderr, "<RFLOW>");
		break;
	case XLOC:
		fprintf(stderr, "<XLOC>");
		break;
	case ENVIRON:
		fprintf(stderr, "<ENVIRON>");
		break;
	case AUTH:
		fprintf(stderr, "<AUTH>");
		break;
	default:
		fprintf(stderr, "<%d>", option);
		break;
	}
	fprintf(stderr, "\n");

	char packet[3] = { TELNET_IAC, state, option };
	TCPLooper::SendRemote(packet, sizeof(packet));
}

#define GetWantState(option) want_state[option]

int TelnetLooper::GetState(option_t option)
{
	return options[option];
}

void TelnetLooper::SetWantState(option_t option, int state)
{
	// TODO: This can almost certainly be done better, sleep on it and come back.
	want_state[option] |= (int)state;
	switch (state) {
	case DO:
		want_state[option] &= ~DONT;
		want_state[option] |= DO;
		break;
	case DONT:
		want_state[option] &= ~DO;
		want_state[option] |= DONT;
		break;
	case WILL:
		want_state[option] &= ~WONT;
		want_state[option] |= WILL;
		break;
	case WONT:
		want_state[option] &= ~WILL;
		want_state[option] |= WONT;
		break;
	}

	// Assume a change in state, I should really check, but it's been
	// a long day and this isn't much network suck.
	switch (state) {
	case WILL:
		SetHisState(option, TELNET_WILL);
		break;
	case WONT:
		SetHisState(option, TELNET_WONT);
		break;
	case DO:
		SetHisState(option, TELNET_DO);
		break;
	case DONT:
		SetHisState(option, TELNET_DONT);
		break;
	}
}

void TelnetLooper::SetState(option_t option, int state)
{
	// TODO: Debugging stuff, clean me out when you can.
	fprintf(stderr, "Recv: <IAC>");
	switch (state) {
	case TELNET_WILL:
		fprintf(stderr, "<WILL>");
		break;
	case TELNET_WONT:
		fprintf(stderr, "<WONT>");
		break;
	case TELNET_DO:
		fprintf(stderr, "<DO>");
		break;
	case TELNET_DONT:
		fprintf(stderr, "<DONT>");
		break;
	default:
		fprintf(stderr, "<%d>", state);
	}
	switch (option) {
	case XMIT_BIN:
		fprintf(stderr, "<XMIT_BIN>");
		break;
	case ECHO:
		fprintf(stderr, "<ECHO>");
		break;
	case NO_GA:
		fprintf(stderr, "<NO_GA>");
		break;
	case STATUS:
		fprintf(stderr, "<STATUS>");
		break;
	case TM:
		fprintf(stderr, "<TM>");
		break;
	case TERM:
		fprintf(stderr, "<TERM>");
		break;
	case EOR:
		fprintf(stderr, "<EOR>");
		break;
	case NAWS:
		fprintf(stderr, "<NAWS>");
		break;
	case TTYSPEED:
		fprintf(stderr, "<TTYSPEED>");
		break;
	case RFLOW:
		fprintf(stderr, "<RFLOW>");
		break;
	case XLOC:
		fprintf(stderr, "<XLOC>");
		break;
	case ENVIRON:
		fprintf(stderr, "<ENVIRON>");
		break;
	case AUTH:
		fprintf(stderr, "<AUTH>");
		break;
	default:
		fprintf(stderr, "<%d>", option);
		break;
	}
	fprintf(stderr, "\n");

	if (options[option] == REQUESTED) {
		/*
		 * We've requested this option, so this must be the ACK. We set
		 * our options to reflect the remote ends reply.
		 */
		switch (state) {
		case TELNET_WILL:
			switch (option) {
			case ECHO:
				options[ECHO] |= DO;
				options[ECHO] &= ~(DONT|WILL);
			default:
				options[option] |= DO;
				options[option] &= ~DONT;
			}
			break;
		case TELNET_WONT:
			options[option] |= DONT;
			options[option] &= ~DO;
			break;
		case TELNET_DO:
			options[option] |= WILL;
			options[option] &= ~WONT;
			break;
		case TELNET_DONT:
			options[option] |= WONT;
			options[option] &= ~WILL;
			break;
		}
		return;
	}

	/*
	 * At this point, the remote end is offering up options that we
	 * haven't requested. So we allow or deny the option and ACK it.
	 */
	int my_state;
	switch (state) {
	case TELNET_DO:
		if (GetWantState(option) & WILL) {
			my_state = TELNET_WILL;
			options[option] |= WILL; options[option] &= ~WONT;
		} else {
			my_state = TELNET_WONT;
			options[option] |= WONT; options[option] &= ~WILL;
		}
		break;
	case TELNET_WILL:
		if (GetWantState(option) & DO) {
			switch (option) {
			case ECHO:
				// Some options only allow one end to use them, so we take
				// care of that possibility here.
				options[option] |= WONT; options[option] &= ~WILL;
			default:
				my_state = TELNET_DO;
				options[option] |= DO; options[option] &= ~DONT;
			}
		} else {
			my_state = TELNET_DONT;
			options[option] |= DONT; options[option] &= ~DO;
		}
		break;
	case TELNET_WONT:
		my_state = TELNET_DONT;
		options[option] |= DONT; options[option] &= ~DO;
		break;
	case TELNET_DONT:
		my_state = TELNET_WONT;
		options[option] |= WONT; options[option] &= ~WILL;
		break;
	}
	SendOption(option, my_state);
}

void TelnetLooper::SetHisState(option_t option, int state)
{
	// Mark the option as requested
	options[option] = REQUESTED;

	SendOption(option, state);
}

TelnetLooper::TelnetLooper(): TCPLooper()
{
	parsing_commands = FALSE;
	for (int i=0; i<TELNET_MAXOPT; i++)
		options[i] = WONT | DONT;

}

void TelnetLooper::InitConnection()
{
	SetWantState(XMIT_BIN, DO);
	SetWantState(ECHO, DO);
	SetWantState(NO_GA, DO);
	TCPLooper::InitConnection();
}

/*
 * Parse incoming data for NVT or TELNET control information before passing
 * it on to the ouput device.
 */
void TelnetLooper::SendLocal(void *data, int32 nbytes)
{
	// TODO: Handle potential errors on calloc().
	char *stripped_data = (char *)calloc(nbytes, sizeof(char));
	unsigned char *fsck = (unsigned char *)data;
	int32 j=0;

	for (int32 i = 0; i < nbytes; i++) {
		switch (fsck[i]) {
		case TELNET_NUL: break;
		case TELNET_BEL: beep(); break;
		case TELNET_LF: stripped_data[j] = '\n'; j++; break;
		case TELNET_CR: stripped_data[j] = '\r'; j++; break;
		case TELNET_IAC: {
			// We've received a command, parse it up.
			switch (fsck[++i]) {
			case TELNET_IAC:
				stripped_data[j] = TELNET_IAC; j++;
				break;
			case TELNET_SE:
				fprintf(stderr, "Recv: <IAC><SE>\n");
				break;
			case TELNET_SB:
				fprintf(stderr, "Recv: <IAC><SB>\n");
				break;
			case TELNET_WILL:
			case TELNET_WONT:
			case TELNET_DO:
			case TELNET_DONT:
				if (fsck[++i] > TELNET_MAXOPT) {
					// Urk. We haven't even heard of this option, NAK it.
					if (fsck[i-1] == TELNET_DO)
						SendOption((option_t)fsck[i], TELNET_WONT);
					else
						SendOption((option_t)fsck[i], TELNET_DONT);
				} else
					SetState((option_t)fsck[i], fsck[i-1]);
				break;
			default:
				fprintf(stderr, "Recv: <IAC><Unknown token: 0x%x>\n", fsck[i]);
			}
		}; break;
		default:
			stripped_data[j] = fsck[i];
			j++;
		}
	}

	free(data);
	TCPLooper::SendLocal(stripped_data, nbytes);
	free(stripped_data);
}